import pytest
from pytest import approx
import pandas as pd

import calliope
from calliope.test.common.util import build_test_model as build_model


class TestNationalScaleExampleModelSenseChecks:
    def test_group_prod_min(self):
        model = calliope.examples.national_scale(
            scenario='cold_fusion_with_production_share'
        )
        model.run()

        df_carrier_prod = (
            model.get_formatted_array('carrier_prod')
                 .loc[dict(carriers='power')].sum('locs').sum('timesteps')
                 .to_pandas()
        )

        prod_share = (
            df_carrier_prod.loc[['cold_fusion', 'csp']].sum() /
            df_carrier_prod.loc[['ccgt', 'cold_fusion', 'csp']].sum()
        )

        assert prod_share == approx(0.85)

    def test_group_cap_max(self):
        model = calliope.examples.national_scale(
            scenario='cold_fusion_with_capacity_share'
        )
        model.run()

        cap_share = (
            model.get_formatted_array('energy_cap').loc[{'techs': ['cold_fusion', 'csp']}].sum() /
            model.get_formatted_array('energy_cap').loc[{'techs': ['ccgt', 'cold_fusion', 'csp']}].sum()
        )

        assert cap_share == approx(0.2)

    def test_systemwide_equals(self):
        model = calliope.examples.national_scale(
            override_dict={
                'techs.ccgt.constraints.energy_cap_max_systemwide': 10000,
                'techs.ac_transmission.constraints.energy_cap_equals_systemwide': 6000
            }
        )
        model.run()
        # Check that setting `_equals` to a finite value leads to forcing
        assert (
            model.get_formatted_array('energy_cap').loc[{'techs': 'ccgt'}].sum() == 10000
        )
        assert (
            model.get_formatted_array('energy_cap').loc[{'techs': 'ac_transmission:region1'}].sum() == 6000
        )

    def test_reserve_margin(self):
        model = calliope.examples.national_scale(
            scenario='reserve_margin'
        )

        model.run()

        # constraint_string = '-Inf : -1.1 * ( carrier_con[region1::demand_power::power,2005-01-05 16:00:00] + carrier_con[region2::demand_power::power,2005-01-05 16:00:00] ) / timestep_resolution[2005-01-05 16:00:00] - energy_cap[region1::ccgt] - energy_cap[region1-3::csp] - energy_cap[region1-2::csp] - energy_cap[region1-1::csp] :   0.0'

        # FIXME: capture Pyomo's print output...
        # assert constraint_string in model._backend_model.reserve_margin_constraint.pprint()

        assert float(model.results.cost.sum()) == approx(282487.35489)


@pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
class TestUrbanScaleMILP:
    def test_asynchronous_prod_con(self):

        def _get_prod_con(model, prod_con):
            return (
                model.get_formatted_array('carrier_{}'.format(prod_con))
                     .loc[{'techs': 'heat_pipes:X1', 'carriers': 'heat'}]
                     .to_pandas().dropna(how='all')
            )
        m = calliope.examples.urban_scale(override_dict={'run.zero_threshold': 1e-6})
        m.run()
        _prod = _get_prod_con(m, 'prod')
        _con = _get_prod_con(m, 'con')
        assert any(((_con < 0) & (_prod > 0)).any()) is True

        m_bin = calliope.examples.urban_scale(
            override_dict={'techs.heat_pipes.constraints.force_asynchronous_prod_con': True,
                           'run.solver_options.mipgap': 0.05,
                           'run.zero_threshold': 1e-6}
        )
        m_bin.run()
        _prod = _get_prod_con(m_bin, 'prod')
        _con = _get_prod_con(m_bin, 'con')
        assert any(((_con < 0) & (_prod > 0)).any()) is False


class TestModelSettings:
    def test_feasibility(self):

        def override(feasibility, cap_val):
            override_dict = {
                'locations.0.techs': {'test_supply_elec': {}, 'test_demand_elec': {}},
                'links.0,1.exists': False,
                # pick a time subset where demand is uniformally -10 throughout
                'model.subset_time': ['2005-01-01 06:00:00', '2005-01-01 08:00:00'],
                'run.ensure_feasibility': feasibility, 'run.bigM': 1e3,
                # Allow setting resource and energy_cap_max/equals to force infeasibility
                'techs.test_supply_elec.constraints': {
                    'resource': cap_val, 'energy_eff': 1, 'energy_cap_equals': 15,
                    'force_resource': True
                }
            }

            return override_dict

        # Feasible case, unmet_demand/unused_supply is deleted
        model_10 = build_model(
            override_dict=override(True, 10),
            scenario='investment_costs'
        )
        model_10.run()
        for i in ['unmet_demand', 'unused_supply']:
            assert hasattr(model_10._backend_model, i)
            assert i not in model_10._model_data.data_vars.keys()

        # Infeasible case, unmet_demand is required
        model_5 = build_model(
            override_dict=override(True, 5),
            scenario='investment_costs'
        )
        model_5.run()
        assert hasattr(model_5._backend_model, 'unmet_demand')
        assert hasattr(model_5._backend_model, 'unused_supply')
        assert model_5._model_data['unmet_demand'].sum() == 15
        assert 'unused_supply' not in model_5._model_data.data_vars.keys()

        # Infeasible case, unused_supply is required
        model_15 = build_model(
            override_dict=override(True, 15),
            scenario='investment_costs'
        )
        model_15.run()
        assert hasattr(model_15._backend_model, 'unmet_demand')
        assert hasattr(model_15._backend_model, 'unused_supply')
        assert model_15._model_data['unmet_demand'].sum() == -15
        assert 'unused_supply' not in model_15._model_data.data_vars.keys()

        assert (
            model_15._backend_model.obj.expr() - model_10._backend_model.obj.expr() ==
            approx(1e3 * 15)
        )

        assert (
            model_5._backend_model.obj.expr() - model_10._backend_model.obj.expr() ==
            approx(1e3 * 15)
        )

        # Infeasible cases = non-optimal termination
        # too much supply
        model = build_model(
            override_dict=override(False, 15),
            scenario='investment_costs'
        )
        model.run()
        assert not hasattr(model._backend_model, 'unmet_demand')
        assert not hasattr(model._backend_model, 'unused_supply')
        assert not model._model_data.attrs['termination_condition'] == 'optimal'

        # too little supply
        model = build_model(
            override_dict=override(False, 5),
            scenario='investment_costs'
        )
        model.run()
        assert not model._model_data.attrs['termination_condition'] == 'optimal'


class TestEnergyCapacityPerStorageCapacity:

    @pytest.fixture
    def model_file(self):
        return "energy_cap_per_storage_cap.yaml"

    @pytest.mark.filterwarnings("ignore:(?s).*`energy_cap_per_storage_cap_min/max/equals`:calliope.exceptions.ModelWarning")
    def test_no_constraint_set(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = model.get_formatted_array("energy_cap").loc[{'techs': 'my_storage'}].sum().item()
        storage_capacity = model.get_formatted_array("storage_cap").loc[{'techs': 'my_storage'}].sum().item()
        assert energy_capacity == pytest.approx(10)
        assert storage_capacity == pytest.approx(180)
        assert storage_capacity != pytest.approx(1 / 10 * energy_capacity)

    def test_equals(self, model_file):
        model = build_model(model_file=model_file, scenario="equals")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = model.get_formatted_array("energy_cap").loc[{'techs': 'my_storage'}].sum().item()
        storage_capacity = model.get_formatted_array("storage_cap").loc[{'techs': 'my_storage'}].sum().item()
        assert storage_capacity == pytest.approx(1 / 10 * energy_capacity)

    def test_max(self, model_file):
        model = build_model(model_file=model_file, scenario="max")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = model.get_formatted_array("energy_cap").loc[{'techs': 'my_storage'}].sum().item()
        storage_capacity = model.get_formatted_array("storage_cap").loc[{'techs': 'my_storage'}].sum().item()
        assert energy_capacity == pytest.approx(10)
        assert storage_capacity == pytest.approx(1000)

    def test_min(self, model_file):
        model = build_model(model_file=model_file, scenario="min")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = model.get_formatted_array("energy_cap").loc[{'techs': 'my_storage'}].sum().item()
        storage_capacity = model.get_formatted_array("storage_cap").loc[{'techs': 'my_storage'}].sum().item()
        assert energy_capacity == pytest.approx(180)
        assert storage_capacity == pytest.approx(180)

    def test_operate_mode(self, model_file):
        model = build_model(model_file=model_file, scenario="operate_mode_min")
        with pytest.raises(calliope.exceptions.ModelError):
            model.run()



import pytest  # pylint: disable=unused-import
import pandas as pd

import calliope
import calliope.exceptions as exceptions

from calliope.test.common.util import build_test_model as build_model
from calliope.test.common.util import check_error_or_warning


@pytest.fixture(scope='class')
def model():
    m = build_model({}, 'simple_supply,two_hours,investment_costs')
    m.run()
    return m


class TestGetInputParams:
    def test_get_input_params(self, model):
        """
        Test that the function access_model_inputs works
        """

        inputs = model.backend.access_model_inputs()
        assert (
            set(model.inputs.data_vars).symmetric_difference(inputs.data_vars)
            == {'objective_cost_class'}
        )


class TestUpdateParam:
    def test_update_param_single_dim(self, model):
        """
        test that the function update_param works with a single dimension
        """

        model.backend.update_param('energy_cap_max', {'1::test_supply_elec': 20})

        assert (
            model._backend_model.energy_cap_max.extract_values()['1::test_supply_elec'] == 20
        )

    def test_update_param_multiple_vals(self, model):
        """
        test that the function update_param works with a single dimension
        """

        model.backend.update_param(
            'energy_cap_max', {'1::test_supply_elec': 20, '0::test_supply_elec': 30}
        )

        assert (
            model._backend_model.energy_cap_max.extract_values()['1::test_supply_elec'] == 20
        )
        assert (
            model._backend_model.energy_cap_max.extract_values()['0::test_supply_elec'] == 30
        )

    def test_update_param_multiple_dim(self, model):
        """
        test that the function update_param works with multiple dimensions
        """

        model.backend.update_param(
            'resource', {('0::test_demand_elec', pd.Timestamp('2005-01-01 01:00:00')): -10}
        )

        assert (
            model._backend_model.resource.extract_values()
            [('0::test_demand_elec', pd.Timestamp('2005-01-01 01:00:00'))] == -10
        )

    def test_unknown_param(self, model):
        """
        Raise error on unknown param
        """

        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.update_param(
                'unknown_param', {('1::test_supply_elec'): 20}
            )

        assert check_error_or_warning(excinfo, 'Parameter `unknown_param` not in the Pyomo Backend.')

    def test_not_a_param(self, model):
        """
        Raise error when trying to update a non-Param Pyomo object
        """
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.update_param(
                'energy_cap', {'1::test_supply_elec': 20}
            )

        assert check_error_or_warning(excinfo, '`energy_cap` not a Parameter in the Pyomo Backend.')

        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.update_param(
                'loc_techs', {'1::test_supply_elec': 20}
            )

        assert check_error_or_warning(excinfo, '`loc_techs` not a Parameter in the Pyomo Backend.')

    def index_not_in_param(self, model):
        """
        Raise error when accessing unknown index
        """

        with pytest.raises(KeyError, match=r"Index 'region1-xc1::csp'"):
            model.backend.update_param(
                'energy_cap_max', {'2::test_supply_elec': 20}
            )

        with pytest.raises(KeyError, match=r"Index 'region1-xc1::csp'"):
            model.backend.update_param(
                'energy_cap_max', {'1::test_supply_elec': 20, '2::test_supply_elec': 20}
            )


class TestActivateConstraint:

    def test_activate_constraint(self, model):
        """
        test that the function activate_constraint works
        """
        model.backend.activate_constraint('system_balance_constraint', active=False)
        assert not model._backend_model.system_balance_constraint.active

        model.backend.activate_constraint('system_balance_constraint', active=True)
        assert model._backend_model.system_balance_constraint.active

    def test_fail_on_activate_unknown_constraint(self, model):
        """
        test that the function activate_constraint fails if unknown constraint
        """
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.activate_constraint('unknown_constraint', active=False)

        assert check_error_or_warning(
            excinfo, 'constraint/objective `unknown_constraint` not in the Pyomo Backend.'
        )

    def test_fail_on_parameter_activate(self, model):
        """
        test that the function activate_constraint fails if trying to activate a
        non-constraint Pyomo object.
        """
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.activate_constraint('resource', active=False)

        assert check_error_or_warning(
            excinfo, '`resource` not a constraint in the Pyomo Backend.'
        )

    def test_non_boolean_parameter_activate(self, model):
        """
        test that the function activate_constraint fails when setting active to
        non-boolean
        """
        with pytest.raises(ValueError) as excinfo:
            model.backend.activate_constraint('system_balance_constraint', active=None)

        assert check_error_or_warning(excinfo, 'Argument `active` must be True or False')


class TestBackendRerun:

    def test_rerun(self, model):
        """
        test that the function rerun works
        """
        with pytest.warns(exceptions.ModelWarning) as excinfo:
            new_model = model.backend.rerun()

        assert isinstance(new_model, calliope.Model)
        for i in ['_timings', 'inputs', 'results']:
            assert hasattr(new_model, i)

        # In new_model.inputs all NaN values have been replace by their default,
        # so we can't directly compare new_model.inputs and model.inputs
        assert new_model.inputs.equals(model.backend.access_model_inputs().reindex(new_model.inputs.coords))

        assert new_model.results.equals(model.results)

        assert check_error_or_warning(
            excinfo,
            'The results of rerunning the backend model are only available'
        )

    def test_update_and_rerun(self, model):
        """
        test that the function rerun works
        """
        model.backend.update_param('energy_cap_max', {'1::test_supply_elec': 20})
        with pytest.warns(exceptions.ModelWarning) as excinfo:
            new_model = model.backend.rerun()

        assert new_model.inputs.energy_cap_max.loc[{'loc_techs': '1::test_supply_elec'}] == 20

        assert check_error_or_warning(
            excinfo,
            'The results of rerunning the backend model are only available'
        )

    def test_rerun_fail_on_operate(self, model):
        # should fail if the run mode is not 'plan'
        model.run_config['mode'] = 'operate'
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.rerun()
        assert check_error_or_warning(
            excinfo, 'Cannot rerun the backend in operate run mode'
        )



import os

import pytest
from pytest import approx
import pandas as pd


import calliope
from calliope.core.attrdict import AttrDict
from calliope.test.common.util import build_test_model as build_model


def get_supply_conversion_techs(model):
    available_techs = [
        'elec_to_heat_cool_unlinked', 'cheap_elec_supply', 'elec_to_heat_cool_linked',
        'elec_to_heat', 'expensive_elec_supply', 'normal_elec_supply',
        'cheap_heat_supply', 'cheap_cool_supply', 'normal_heat_supply',
        'expensive_heat_supply', 'normal_cool_supply', 'expensive_cool_supply',
        'elec_supply_plus'
    ]
    return [i for i in available_techs if i in model._model_data.techs.values]


@pytest.fixture(scope='module')
def model_file():
    return os.path.join('model_config_group', 'base_model.yaml')


# Group constraints, i.e. those that can be defined on a system/subsystem scale
@pytest.mark.filterwarnings("ignore:(?s).*All technologies were requested:calliope.exceptions.ModelWarning")
class TestBuildGroupConstraints:

    metadata_yaml = os.path.join(
        os.path.dirname(__file__), 'common', 'test_model',
        'model_config_group', 'scenarios_metadata.yaml'
    )

    scenario_metadata = AttrDict.from_yaml(metadata_yaml)
    test_vars = [
        (k, v['constraint'], v['group_name'], v['loc_techs'])
        for k, v in scenario_metadata.items()
    ]

    @pytest.mark.parametrize("scenario, constraints, group_names, loc_techs", test_vars)
    def test_build_group_constraint(self, scenario, constraints, group_names, loc_techs):

        scenario_model = build_model(
            model_file=os.path.join('model_config_group', 'base_model.yaml'),
            scenario=scenario
        )

        assert all('group_names_' + i in scenario_model._model_data.dims for i in constraints)
        assert all(
            group_names[i] in scenario_model._model_data['group_names_' + constraints[i]].values
            for i in range(len(constraints))
        )
        assert all('group_' + i in scenario_model._model_data.data_vars.keys() for i in constraints)
        assert all(
            len(set(loc_techs[i]) -
                set(scenario_model._model_data['group_constraint_loc_techs_' + group_names[i]].values))
            == 0 for i in range(len(group_names))
        )

        scenario_model.run(build_only=True)
        assert all(
            hasattr(scenario_model._backend_model, 'group_' + i + '_constraint') for i in constraints
        )

@pytest.mark.xfail(reason="Tests not yet implemented.")
@pytest.mark.filterwarnings("ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning")
class TestGroupConstraints:

    def test_no_group_constraint(self):
        model = build_model(model_file="group_constraints.yaml")
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
                 .loc[{'techs': 'expensive_elec_supply'}].sum().item()
        )
        assert expensive_generation == 0

    def test_switched_off_group_constraint(self):
        model = build_model(
            model_file="group_constraints.yaml",
            scenario="switching_off_group_constraint"
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
                 .loc[{'techs': 'expensive_elec_supply'}].sum().item()
        )
        assert expensive_generation == 0

    @pytest.mark.xfail(reason="Check not yet implemented.")
    def test_group_constraint_without_technology(self):
        model = build_model(
            model_file='group_constraints.yaml',
            scenario='group_constraint_without_tech'
        )
        with pytest.raises(calliope.exceptions.ModelError):
            model.run()

    def test_group_constraint_with_several_constraints(self):
        model = build_model(
            model_file='group_constraints.yaml',
            scenario='several_group_constraints'
        )
        model.run()
        expensive_generation = (model.get_formatted_array("carrier_prod")
                                     .to_dataframe()
                                     .reset_index()
                                     .groupby("techs")
                                     .carrier_prod
                                     .sum()
                                     .transform(lambda x: x / x.sum())
                                     .loc["expensive_elec_supply"])
        assert round(expensive_generation, 5) >= 0.8


@pytest.mark.xfail(reason="Tests not yet implemented.")
@pytest.mark.filterwarnings("ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning")
class TestDemandShareGroupConstraints:
    def test_no_demand_share_constraint(self):
        model = build_model(model_file='demand_share.yaml')
        model.run()
        expensive_generation = (model.get_formatted_array("carrier_prod")
                                     .to_dataframe()
                                     .reset_index()
                                     .groupby("techs")
                                     .carrier_prod
                                     .sum()
                                     .loc["expensive_elec_supply"])
        assert expensive_generation == 0

    def test_systemwide_demand_share_max_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_max_systemwide'
        )
        model.run()
        cheap_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "cheap_elec_supply", 'carriers': "electricity"}].sum()
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (cheap_generation / demand).round(5) <= 0.3

    def test_systemwide_demand_share_min_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_min_systemwide'
        )
        model.run()
        expensive_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum()
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (expensive_generation / demand).round(5) >= 0.6

    def test_systemwide_demand_share_equals_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_equals_systemwide'
        )
        model.run()
        expensive_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum()
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (expensive_generation / demand).round(5) == 0.6

    def test_location_specific_demand_share_max_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_max_location_0'
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(dim='timesteps').loc[{'carriers': "electricity"}]
        demand0 = -model.get_formatted_array("carrier_con").loc[{'locs': '0'}].sum().item()
        cheap_generation0 = generation.loc[{'locs': "0", 'techs': "cheap_elec_supply"}].item()
        expensive_generation1 = generation.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_generation0 / demand0, 5) <= 0.3
        assert expensive_generation1 == 0

    def test_location_specific_demand_share_max_constraint_two_techs(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_max_location_0_two_techs'
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(dim='timesteps').loc[{'carriers': "electricity"}]
        demand0 = -model.get_formatted_array("carrier_con").loc[{'locs': '0'}].sum().item()
        generation0 = generation.loc[{'locs': "0", 'techs': ["cheap_elec_supply", "normal_elec_supply"]}].sum('techs').item()
        assert round(generation0 / demand0, 5) <= 0.4

    def test_location_specific_demand_share_min_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_min_location_0'
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(dim='timesteps').loc[{'carriers': "electricity"}]
        demand0 = -model.get_formatted_array("carrier_con").loc[{'locs': '0'}].sum().item()
        expensive_generation0 = generation.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_generation1 = generation.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(expensive_generation0 / demand0, 5) >= 0.6
        assert expensive_generation1 == 0

    def test_multiple_group_constraints(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='multiple_constraints'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'locs', 'carriers')))
        demand = -model.get_formatted_array("carrier_con").sum().item()
        cheap_generation = generation.loc[{'techs': 'cheap_elec_supply'}].item()
        expensive_generation = generation.loc[{'techs': 'expensive_elec_supply'}].item()

        assert round(expensive_generation / demand, 5) >= 0.6
        assert round(cheap_generation / demand, 5) <= 0.3

    def test_multiple_group_carriers(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='multiple_carriers_max'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'locs')))
        demand = (-model.get_formatted_array("carrier_con")
                        .sum(dim=('timesteps', 'locs')))
        cheap_generation_elec = generation.loc[{'techs': 'cheap_elec_supply', 'carriers': 'electricity'}].item()
        demand_elec = demand.loc[{'techs': 'electricity_demand', 'carriers': 'electricity'}].item()
        cheap_generation_heat = generation.loc[{'techs': 'cheap_heat_supply', 'carriers': 'heat'}].item()
        demand_heat = demand.loc[{'techs': 'heat_demand', 'carriers': 'heat'}].item()

        assert round(cheap_generation_elec / demand_elec, 5) <= 0.3
        assert round(cheap_generation_heat / demand_heat, 5) <= 0.5

    def test_multiple_group_carriers_constraints(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='multiple_constraints_carriers'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'locs')))
        demand = (-model.get_formatted_array("carrier_con")
                        .sum(dim=('timesteps', 'locs')))
        cheap_generation_elec = generation.loc[{'techs': 'cheap_elec_supply', 'carriers': 'electricity'}].item()
        expensive_generation_elec = generation.loc[{'techs': 'expensive_elec_supply', 'carriers': 'electricity'}].item()
        demand_elec = demand.loc[{'techs': 'electricity_demand', 'carriers': 'electricity'}].item()
        cheap_generation_heat = generation.loc[{'techs': 'cheap_heat_supply', 'carriers': 'heat'}].item()
        expensive_generation_heat = generation.loc[{'techs': 'expensive_heat_supply', 'carriers': 'heat'}].item()
        demand_heat = demand.loc[{'techs': 'heat_demand', 'carriers': 'heat'}].item()

        assert round(cheap_generation_elec / demand_elec, 5) <= 0.3
        assert round(expensive_generation_elec / demand_elec, 5) >= 0.6
        assert round(cheap_generation_heat / demand_heat, 5) <= 0.5
        assert round(expensive_generation_heat / demand_heat, 5) >= 0.4

    def test_different_locations_per_group_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='different_locations_per_group'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'carriers')))
        demand = (-model.get_formatted_array("carrier_con")
                        .sum(dim=('timesteps')))
        cheap_generation_0 = generation.loc[{'techs': 'cheap_elec_supply', 'locs': '0'}].item()
        expensive_generation_0 = generation.loc[{'techs': 'expensive_elec_supply', 'locs': '0'}].item()
        cheap_generation_1 = generation.loc[{'techs': 'cheap_elec_supply', 'locs': '1'}].item()
        expensive_generation_1 = generation.loc[{'techs': 'expensive_elec_supply', 'locs': '1'}].item()
        demand_elec_0 = demand.loc[{'techs': 'electricity_demand', 'carriers': 'electricity', 'locs': '0'}].item()
        demand_elec_1 = demand.loc[{'techs': 'electricity_demand', 'carriers': 'electricity', 'locs': '1'}].item()

        assert round(expensive_generation_0 / demand_elec_0, 5) >= 0.6
        assert expensive_generation_1 / demand_elec_1 == 0
        assert round((cheap_generation_0 + cheap_generation_1) / (demand_elec_0 + demand_elec_1), 5) <= 0.3

    def test_transmission_not_included_in_demand(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='transmission_not_included_in_demand'
        )
        model.run()
        assert model.results.termination_condition == "optimal"
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'carriers')))
        demand = (-model.get_formatted_array("carrier_con")
                        .sum(dim=('timesteps', 'carriers')))

        assert (generation.sel(locs="1", techs=["normal_elec_supply", "cheap_elec_supply", "expensive_elec_supply"])
                          .sum(dim="techs")
                          .item()) == approx(0)

        cheap_elec_supply_0 = generation.sel(locs="0", techs="cheap_elec_supply").item()
        demand_0 = demand.sel(locs="0", techs="electricity_demand").item()

        assert round(cheap_elec_supply_0 / demand_0, 5) <= 0.4

    def test_demand_share_per_timestep_max(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_per_timestep_max'
        )
        model.run()
        cheap_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "cheap_elec_supply", 'carriers': "electricity"}].sum('locs')
        demand = -1 * model.get_formatted_array("carrier_con").sum('locs')
        # assert share in each timestep is 0.6
        assert ((cheap_generation / demand).round(5) <= 0.3).all()

    def test_demand_share_per_timestep_min(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_per_timestep_min'
        )
        model.run()
        expensive_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum('locs')
        demand = -1 * model.get_formatted_array("carrier_con").sum('locs')
        # assert share in each timestep is 0.6
        assert ((expensive_generation / demand).round(5) >= 0.6).all()

    def test_demand_share_per_timestep_equals(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_per_timestep_equals'
        )
        model.run()
        expensive_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum('locs')
        demand = -1 * model.get_formatted_array("carrier_con").sum('locs')
        # assert share in each timestep is 0.6
        assert ((expensive_generation / demand).round(5) == 0.6).all()

    def test_demand_share_per_timestep_decision_inf(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('locs').sum('techs').to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert (shares['cheap_elec_supply'] == 0.1875).all()
        assert (shares['normal_elec_supply'] == 0.8125).all()

    def test_demand_share_per_timestep_decision_simple(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_simple'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('locs').sum('techs').to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert (shares['cheap_elec_supply'] == 0.1875).all()
        assert (shares['normal_elec_supply'] == 0.8125).all()

    def test_demand_share_per_timestep_decision_not_one(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_not_one'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('locs').sum('techs').to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert (shares['cheap_elec_supply'] == 0.1).all()
        assert (shares['normal_elec_supply'] == 0.9).all()
        assert shares['expensive_elec_supply'].mean() + shares['normal_elec_supply'].mean() == 0.9

    def test_demand_share_per_timestep_decision_per_location(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_per_location'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('techs').to_series()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].to_series()
        shares = supply.div(demand, axis=0)

        assert (shares.loc[pd.IndexSlice['0', :, :]].unstack()['cheap_elec_supply'] == 0.25).all()
        assert (shares.loc[pd.IndexSlice['0', :, :]].unstack()['normal_elec_supply'] == 0.75).all()

        assert (shares.loc[pd.IndexSlice['1', :, :]].unstack()['cheap_elec_supply'] == 0.125).all()
        assert (shares.loc[pd.IndexSlice['1', :, :]].unstack()['normal_elec_supply'] == 0.875).all()

    def test_demand_share_per_timestep_decision_inf_with_transmission(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf,with_electricity_transmission'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('techs').to_series()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].to_series()
        shares = supply.div(demand, axis=0)

        assert all([i == approx(0.12373737) for i in shares.loc[pd.IndexSlice['0', :, :]].unstack()['cheap_elec_supply'].values])
        assert all([i == approx(0.87626263) for i in shares.loc[pd.IndexSlice['0', :, :]].unstack()['normal_elec_supply'].values])
        assert (shares.loc[pd.IndexSlice['1', :, :]].unstack()['electricity_transmission:0'] == 1.0).all()

    def test_demand_share_per_timestep_decision_inf_with_heat_cosntrain_electricity(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf,with_electricity_conversion_tech'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum(['locs', 'techs']).to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert all([i == approx(0.175926) for i in shares['cheap_elec_supply'].values])
        assert all([i == approx(0.824074) for i in shares['normal_elec_supply'].values])

    def test_demand_share_per_timestep_decision_inf_with_heat_constrain_heat(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf_with_heat,with_electricity_conversion_tech'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "heat"}].sum(['locs', 'techs']).to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "heat"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert all([i == approx(0.5) for i in shares['elec_to_heat'].values])
        assert all([i == approx(0.5) for i in shares['heating'].values])

    def test_demand_share_per_timestep_decision_inf_with_heat_constrain_heat_and_electricity(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf_with_heat,demand_share_per_timestep_decision_not_one,with_electricity_conversion_tech'
        )
        model.run()
        demand_heat = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "heat"}].sum(['locs', 'techs']).to_pandas()
        supply_heat = model.get_formatted_array("carrier_prod").loc[{'carriers': "heat"}].sum('locs').to_pandas().T
        shares_heat = supply_heat.div(demand_heat, axis=0)

        demand_elec = -1 * (
            model._model_data.carrier_con.loc[{
                'loc_tech_carriers_con': [
                    i for i in model._model_data.loc_tech_carriers_demand.values
                    if 'electricity' in i
                ]
            }].sum(['loc_tech_carriers_con']).to_pandas()
        )
        supply_elec = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares_elec = supply_elec.div(demand_elec, axis=0)

        assert all([i == approx(0.5) for i in shares_heat['elec_to_heat'].values])
        assert all([i == approx(0.5) for i in shares_heat['heating'].values])
        assert all([i == approx(0.9) for i in shares_elec['normal_elec_supply'].values])
        assert all([round(i, 5) >= 0.1 for i in shares_elec['cheap_elec_supply'].values])


@pytest.mark.filterwarnings("ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning")
class TestResourceAreaGroupConstraints:

    def test_no_resource_area_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        capacity = model.get_formatted_array("resource_area")
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert expensive_capacity == 0

    def test_resource_area_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='resource_area_max_supply'
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        cheap_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(cheap_capacity, 5) <= 28

    def test_resource_area_min_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='resource_area_min_supply'
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(expensive_capacity, 5) >= 12

    def test_resource_area_equals_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='resource_area_equals_supply'
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert expensive_capacity == approx(20)

    def test_resource_area_min_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='resource_area_min_max_supply'
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        cheap_capacity = capacity.loc[{'techs': "cheap_elec_supply"}].sum().item()
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(cheap_capacity, 5) <= 8
        assert round(expensive_capacity, 5) >= 12

    def test_resource_area_max_supply_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='resource_area_max_supply_loc_1'
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        cheap_capacity1 = capacity.loc[{'locs': "1", 'techs': "cheap_elec_supply"}].item()
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_capacity1, 5) <= 8
        assert expensive_capacity0 == 0

    def test_resource_area_min_supply_loc_0_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='resource_area_min_supply_loc_0'
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(expensive_capacity0, 5) >= 12
        assert expensive_capacity1 == 0

    def test_resource_area_min_max_supply_loc0_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='resource_area_min_max_supply_loc_0_1'
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        cheap_capacity1 = capacity.loc[{'locs': "1", 'techs': "cheap_elec_supply"}].item()
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_capacity1, 5) <= 8
        assert round(expensive_capacity0, 5) >= 12
        assert expensive_capacity1 == 0

    # All technologies, but insufficient resource_area_max for enough installed capacity to meet demand
    @pytest.mark.filterwarnings(
        "ignore:(?s).*['conversion', 'conversion_plus', 'demand', 'storage', "
        "'transmission'].*:calliope.exceptions.ModelWarning"
    )
    def test_resource_area_max_all_techs_infeasible_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='resource_area_max_all_techs_infeasible'
        )

        model.run()

        assert model._model_data.attrs['termination_condition'] != 'optimal'


@pytest.mark.xfail(reason="Tests not yet implemented.")
@pytest.mark.filterwarnings("ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning")
class TestCostCapGroupConstraint:

    def test_systemwide_cost_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_max_systemwide'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        assert round(cheap_cost, 5) <= 30

    def test_systemwide_cost_investment_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_investment_max_systemwide'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        cheap_cost_investment = (model.get_formatted_array('cost_investment')
                                      .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        assert cheap_cost > cheap_cost_investment
        assert round(cheap_cost_investment, 5) <= 4

    def test_systemwide_cost_var_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_var_max_systemwide'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        cheap_cost_var = (model.get_formatted_array('cost_var')
                               .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        assert cheap_cost > cheap_cost_var
        assert round(cheap_cost_var, 5) <= 200

    def test_systemwide_cost_min_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='expensive_cost_min_systemwide'
        )
        model.run()
        expensive_cost = (model.get_formatted_array('cost')
                               .loc[{'costs': 'monetary', 'techs': 'expensive_clean_supply'}]).sum().item()
        assert round(expensive_cost, 5) >= 600

    def test_systemwide_cost_equals_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_equals_systemwide'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        assert cheap_cost == approx(210)

    def test_location_specific_cost_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_max_location_0'
        )
        model.run()
        cheap_cost0 = (model.get_formatted_array('cost')
                            .loc[{'costs': 'monetary',
                                  'techs': 'cheap_polluting_supply',
                                  'locs': '0'}]).sum().item()
        assert round(cheap_cost0, 5) <= 10

    def test_systemwide_emissions_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='emissions_max_systemwide'
        )
        model.run()
        emissions = (model.get_formatted_array('cost')
                          .loc[{'costs': 'emissions'}]).sum().item()
        assert round(emissions, 5) <= 400

    def test_location_specific_emissions_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='emissions_max_location_0'
        )
        model.run()
        emissions0 = (model.get_formatted_array('cost')
                           .loc[{'costs': 'emissions',
                                 'locs': '0'}]).sum().item()
        assert round(emissions0, 5) <= 200

    def test_systemwide_clean_emissions_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='clean_emissions_max_systemwide'
        )
        model.run()
        clean_emissions = (model.get_formatted_array('cost')
                                .loc[{'costs': 'emissions',
                                      'techs': 'expensive_clean_supply'}]).sum().item()
        assert round(clean_emissions, 5) <= 300

    def test_multiple_costs_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='multiple_costs_constraint'
        )
        model.run()
        emissions = (model.get_formatted_array('cost')
                          .loc[{'costs': 'emissions'}]).sum().item()
        expensive_cost = (model.get_formatted_array('cost')
                               .loc[{'costs': 'monetary',
                                     'techs': 'expensive_clean_supply'}]).sum().item()
        assert round(emissions, 5) <= 400
        assert round(expensive_cost, 5) <= 600

    def test_different_locations_per_cost_group_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='different_locations_per_group'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        cheap_cost0 = (model.get_formatted_array('cost')
                            .loc[{'costs': 'monetary',
                                  'techs': 'cheap_polluting_supply',
                                  'locs': '0'}]).sum().item()
        assert round(cheap_cost, 5) <= 30
        assert round(cheap_cost0, 5) <= 10

    def test_different_techs_per_cost_group_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='different_techs_per_group'
        )
        model.run()
        emissions = (model.get_formatted_array('cost')
                          .loc[{'costs': 'emissions'}]).sum().item()
        clean_emissions = (model.get_formatted_array('cost')
                                .loc[{'costs': 'emissions',
                                      'techs': 'expensive_clean_supply'}]).sum().item()
        assert round(emissions, 5) <= 400
        assert round(clean_emissions, 5) <= 300


@pytest.mark.xfail(reason="Tests not yet implemented.")
@pytest.mark.filterwarnings("ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning")
class TestSupplyShareGroupConstraints:

    def test_no_carrier_prod_share_constraint(self):
        model = build_model(model_file='supply_share.yaml')
        model.run()
        expensive_generation = (model.get_formatted_array("carrier_prod")
                                     .loc[{'techs': "expensive_elec_supply"}].sum()).item()
        assert expensive_generation == 0

    def test_systemwide_carrier_prod_share_max_constraint(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_max_systemwide'
        )
        model.run()
        cheap_generation = (model.get_formatted_array("carrier_prod")
                                 .to_dataframe()
                                 .reset_index()
                                 .groupby("techs")
                                 .carrier_prod
                                 .sum()
                                 .transform(lambda x: x / x.sum())
                                 .loc["cheap_elec_supply"])
        assert round(cheap_generation, 5) <= 0.4

    def test_systemwide_carrier_prod_share_min_constraint(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_min_systemwide'
        )
        model.run()
        expensive_generation = (model.get_formatted_array("carrier_prod")
                                     .to_dataframe()
                                     .reset_index()
                                     .groupby("techs")
                                     .carrier_prod
                                     .sum()
                                     .transform(lambda x: x / x.sum())
                                     .loc["expensive_elec_supply"])
        assert round(expensive_generation, 5) >= 0.6

    def test_location_specific_carrier_prod_share_max_constraint(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_max_location_0'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim='timesteps').loc[{'carriers': 'electricity'}])
        cheap_generation0 = generation.loc[{'locs': "0", 'techs': "cheap_elec_supply"}].item()
        expensive_generation0 = generation.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_generation1 = generation.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_generation0 / (cheap_generation0 + expensive_generation0), 5) <= 0.4
        assert expensive_generation1 == 0

    def test_location_specific_carrier_prod_share_min_constraint(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_min_location_0'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim='timesteps').loc[{'carriers': 'electricity'}])
        cheap_generation0 = generation.loc[{'locs': "0", 'techs': "cheap_elec_supply"}].item()
        expensive_generation0 = generation.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_generation1 = generation.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(expensive_generation0 / (cheap_generation0 + expensive_generation0), 5) >= 0.6
        assert expensive_generation1 == 0

    def test_carrier_prod_share_with_transmission(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_min_systemwide,transmission_link'
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
                 .loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        supply = (
            model._model_data.carrier_prod
            .loc[{"loc_tech_carriers_prod":
                  model._model_data.loc_tech_carriers_supply_conversion_all.values}]
            .sum().item()
        )

        assert round(expensive_generation / supply, 5) >= 0.6

    def test_carrier_prod_share_with_storage(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_min_systemwide,storage_tech'
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
                 .loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        supply = (
            model._model_data.carrier_prod
            .loc[{"loc_tech_carriers_prod":
                  model._model_data.loc_tech_carriers_supply_conversion_all.values}]
            .sum().item()
        )

        assert round(expensive_generation / supply, 5) >= 0.6

    def test_carrier_prod_share_per_timestep_max(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_per_timestep_max'
        )
        model.run()
        cheap_elec_supply = model.get_formatted_array("carrier_prod").loc[{'techs': "cheap_elec_supply", 'carriers': "electricity"}].sum('locs')
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').sum('techs')
        # assert share in each timestep is 0.4
        assert ((cheap_elec_supply / supply).round(5) <= 0.4).all()

    def test_carrier_prod_share_per_timestep_min(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_per_timestep_min'
        )
        model.run()
        expensive_elec_supply = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum('locs')
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').sum('techs')
        # assert share in each timestep is 0.6
        assert ((expensive_elec_supply / supply).round(5) >= 0.6).all()

    def test_carrier_prod_share_per_timestep_equals(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='carrier_prod_share_per_timestep_equals'
        )
        model.run()
        expensive_elec_supply = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum('locs')
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').sum('techs')
        # assert share in each timestep is 0.6
        assert ((expensive_elec_supply / supply).round(5) == 0.6).all()


class TestEnergyCapShareGroupConstraints:

    @pytest.fixture(scope='module')
    def model_file(self):
        return os.path.join('model_config_group', 'base_model.yaml')

    def test_no_energy_cap_share_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert expensive_capacity == 0

    def test_energy_cap_share_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_max_supply'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all = (
            capacity.loc[{'techs': get_supply_conversion_techs(model)}]
        )
        cheap_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(cheap_capacity / capacity_supply_conversion_all.sum().item(), 5) <= 0.5

    def test_energy_cap_share_min_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_min_supply'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all = (
            capacity.loc[{'techs': get_supply_conversion_techs(model)}]
        )
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(expensive_capacity / capacity_supply_conversion_all.sum().item(), 5) >= 0.4

    def test_energy_cap_share_equals_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_equals_supply'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all = (
            capacity.loc[{'techs': get_supply_conversion_techs(model)}]
        )
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert expensive_capacity / capacity_supply_conversion_all.sum().item() == approx(0.3)

    def test_energy_cap_share_min_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_min_max_supply'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all = (
            capacity.loc[{'techs': get_supply_conversion_techs(model)}]
        )
        cheap_capacity = capacity.loc[{'techs': "cheap_elec_supply"}].sum().item()
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(cheap_capacity / capacity_supply_conversion_all.sum().item(), 5) <= 0.2
        assert round(expensive_capacity / capacity_supply_conversion_all.sum().item(), 5) >= 0.4

    def test_energy_cap_share_max_supply_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_max_supply_loc_1'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all_loc1 = (
            capacity.loc[{'locs': "1", 'techs': get_supply_conversion_techs(model)}]
        )
        cheap_capacity1 = capacity.loc[{'locs': "1", 'techs': "cheap_elec_supply"}].item()
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_capacity1 / capacity_supply_conversion_all_loc1.sum().item(), 5) <= 0.2
        assert expensive_capacity0 == 0

    def test_energy_cap_share_min_supply_loc_0_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_min_supply_loc_0'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all_loc0 = (
            capacity.loc[{'locs': "0", 'techs': get_supply_conversion_techs(model)}]
        )
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(expensive_capacity0 / capacity_supply_conversion_all_loc0.sum().item(), 5) >= 0.4
        assert expensive_capacity1 == 0

    def test_energy_cap_share_min_max_supply_loc0_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_min_max_supply_loc_0_1'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all_loc0 = (
            capacity.loc[{'locs': "0", 'techs': get_supply_conversion_techs(model)}]
        )
        capacity_supply_conversion_all_loc1 = (
            capacity.loc[{'locs': "1", 'techs': get_supply_conversion_techs(model)}]
        )
        cheap_capacity1 = capacity.loc[{'locs': "1", 'techs': "cheap_elec_supply"}].item()
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_capacity1 / capacity_supply_conversion_all_loc1.sum().item(), 5) <= 0.2
        assert round(expensive_capacity0 / capacity_supply_conversion_all_loc0.sum().item(), 5) >= 0.4
        assert expensive_capacity1 == 0

    # All conversion technologies with insufficient energy_cap_max to use the
    # cheap direct heat/cooling supply techs
    def test_energy_cap_share_max_non_conversion_all_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_max_non_conversion_all'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all_loc1 = (
            capacity.loc[{'locs': "1", 'techs': get_supply_conversion_techs(model)}]
        )
        cheap_heat = capacity.loc[{'locs': "1", 'techs': "cheap_heat_supply"}].item()
        cheap_cool = capacity.loc[{'locs': "1", 'techs': "cheap_cool_supply"}].item()
        assert (cheap_heat + cheap_cool) / capacity_supply_conversion_all_loc1.sum().item() <= 0.1

    # All technologies, but insufficient energy_cap_max for enough installed capacity to meet demand
    @pytest.mark.filterwarnings("ignore:(?s).*`['demand']` have been ignored*:calliope.exceptions.ModelWarning")
    def test_energy_cap_share_max_all_techs_infeasible_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_share_max_all_techs_infeasible'
        )

        model.run()

        assert model._model_data.attrs['termination_condition'] != 'optimal'


class TestEnergyCapGroupConstraints:

    def test_no_energy_cap_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert expensive_capacity == 0

    def test_energy_cap_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_max_supply'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity = capacity.loc[{'techs': "cheap_elec_supply"}].sum().item()
        assert round(cheap_capacity, 5) <= 14

    def test_energy_cap_min_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_min_supply'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(expensive_capacity, 5) >= 6

    def test_energy_cap_equals_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_equals_supply'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert expensive_capacity == approx(10)

    def test_energy_cap_min_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_min_max_supply'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity = capacity.loc[{'techs': "cheap_elec_supply"}].sum().item()
        expensive_capacity = capacity.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(cheap_capacity, 5) <= 4
        assert round(expensive_capacity, 5) >= 6

    def test_energy_cap_max_supply_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_max_supply_loc_1'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity1 = capacity.loc[{'locs': "1", 'techs': "cheap_elec_supply"}].item()
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_capacity1, 5) <= 4
        assert expensive_capacity0 == 0

    def test_energy_cap_min_supply_loc_0_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_min_supply_loc_0'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(expensive_capacity0, 5) >= 6
        assert expensive_capacity1 == 0

    def test_energy_cap_min_max_supply_loc0_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_min_max_supply_loc_0_1'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity1 = capacity.loc[{'locs': "1", 'techs': "cheap_elec_supply"}].item()
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_capacity1, 5) <= 4
        assert round(expensive_capacity0, 5) >= 6
        assert expensive_capacity1 == 0

    # All conversion technologies with insufficient energy_cap_max to use the
    # cheap direct heat/cooling supply techs
    def test_energy_cap_max_non_conversion_all_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_max_non_conversion_all'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        elec_to_heat = capacity.loc[{'locs': "1", 'techs': "elec_to_heat"}].item()
        elec_to_heat_cool_linked = capacity.loc[{'locs': "1", 'techs': "elec_to_heat_cool_linked"}].item()
        elec_to_heat_cool_unlinked = capacity.loc[{'locs': "1", 'techs': "elec_to_heat_cool_unlinked"}].item()
        cheap_heat = capacity.loc[{'locs': "1", 'techs': "cheap_heat_supply"}].item()
        cheap_cool = capacity.loc[{'locs': "1", 'techs': "cheap_cool_supply"}].item()
        assert cheap_heat == approx(0.8333, rel=0.001)
        assert cheap_cool == approx(0.1667, rel=0.001)
        assert elec_to_heat_cool_linked == approx(1.6667, rel=0.001)
        assert elec_to_heat_cool_unlinked == 0
        assert elec_to_heat == 0

    # All technologies, but insufficient energy_cap_max for enough installed capacity to meet demand
    @pytest.mark.filterwarnings("ignore:(?s).*`['demand']` have been ignored*:calliope.exceptions.ModelWarning")
    def test_energy_cap_max_all_techs_infeasible_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='energy_cap_max_all_techs_infeasible'
        )

        model.run()

        assert model._model_data.attrs['termination_condition'] != 'optimal'


class TestNetImportShareGroupConstraints:

    @pytest.fixture
    def results_for_scenario(self, model_file):
        def __run(scenario=None):
            model = build_model(model_file=model_file, scenario=scenario)
            model.run()
            return model
        return __run

    @staticmethod
    def retrieve_imports(results):
        carrier_prod = results.get_formatted_array("carrier_prod").sel(carriers="electricity")
        return (carrier_prod
                .sel(techs=["transmission" in str(tech) for tech in carrier_prod.techs])
                .sum(["techs", "timesteps"]))

    @staticmethod
    def retrieve_exports(results):
        carrier_prod = results.get_formatted_array("carrier_con").sel(carriers="electricity")
        return (carrier_prod
                .sel(techs=["transmission" in str(tech) for tech in carrier_prod.techs])
                .sum(["techs", "timesteps"])) * (-1)

    @staticmethod
    def retrieve_net_imports(results):
        imports = TestNetImportShareGroupConstraints.retrieve_imports(results)
        exports = TestNetImportShareGroupConstraints.retrieve_exports(results)
        return imports - exports

    @staticmethod
    def retrieve_demand(results):
        return (results
                .get_formatted_array("carrier_con")
                .sel(carriers="electricity", techs="electricity_demand")
                .sum(["timesteps"]))

    def test_only_imports_without_constraint(self, results_for_scenario):
        results = results_for_scenario("expensive-1")
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(- demand)

    def test_only_imports_with_wrong_carrier(self, results_for_scenario):
        results = results_for_scenario("expensive-1,other-carrier")
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(- demand)

    def test_no_net_imports(self, results_for_scenario):
        results = results_for_scenario("expensive-1,no-net-imports")
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(0)

    @pytest.mark.xfail(reason="One cannot define transmission techs explicitely.")
    def test_no_imports_explicit_tech(self, results_for_scenario):
        results = results_for_scenario("expensive-1,no-net-imports-explicit-tech")
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(0)

    def test_some_imports_allowed(self, results_for_scenario):
        results = results_for_scenario("expensive-1,some-imports-allowed")
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports <= - 0.2 * demand

    def test_some_imports_enforced(self, results_for_scenario):
        results = results_for_scenario("expensive-1,some-imports-enforced")
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(- 0.2 * demand)

    def test_some_expensive_imports_enforced(self, results_for_scenario):
        results = results_for_scenario("expensive-1,some-expensive-imports-enforced")
        demand = self.retrieve_demand(results).sel(locs="0").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="0").item()
        assert net_imports >= - 0.2 * demand

    def test_ignores_imports_within_group(self, results_for_scenario):
        results = results_for_scenario("expensive-1,ignores-imports-within-group")
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(- demand)

    def test_allows_gross_imports(self, results_for_scenario):
        results = results_for_scenario("no-net-imports,alternating-costs")
        gross_imports = self.retrieve_imports(results).sel(locs="1").item()
        assert gross_imports > 0

    def test_no_net_imports_despite_gross_imports(self, results_for_scenario):
        results = results_for_scenario("no-net-imports,alternating-costs")
        net_imports = self.retrieve_net_imports(results).sel(locs="1")
        assert net_imports <= 0


class TestCarrierProdGroupConstraints:

    def test_no_carrier_prod_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        expensive_prod = prod.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert expensive_prod == 0

    def test_carrier_prod_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_max_supply'
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        cheap_prod = prod.loc[{'techs': "cheap_elec_supply"}].sum().item()
        assert round(cheap_prod, 5) <= 40

    def test_carrier_prod_min_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_min_supply'
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        expensive_prod = prod.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(expensive_prod, 5) >= 10

    def test_carrier_prod_equals_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_equals_supply'
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        expensive_prod = prod.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert expensive_prod == approx(30)

    def test_carrier_prod_min_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_min_max_supply'
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        cheap_prod = prod.loc[{'techs': "cheap_elec_supply"}].sum().item()
        expensive_prod = prod.loc[{'techs': "expensive_elec_supply"}].sum().item()
        assert round(cheap_prod, 5) <= 40
        assert round(expensive_prod, 5) >= 10

    def test_carrier_prod_max_supply_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_max_supply_loc_1'
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        cheap_prod1 = prod.loc[{'locs': "1", 'techs': "cheap_elec_supply"}].sum().item()
        expensive_prod0 = prod.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].sum().item()
        assert round(cheap_prod1, 5) <= 40
        assert expensive_prod0 == 0

    def test_carrier_prod_min_supply_loc_0_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_min_supply_loc_0'
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        expensive_prod0 = prod.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].sum().item()
        expensive_prod1 = prod.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].sum().item()
        assert round(expensive_prod0, 5) >= 10
        assert expensive_prod1 == 0

    def test_carrier_prod_min_max_supply_loc0_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_min_max_supply_loc_0_1'
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        cheap_prod1 = prod.loc[{'locs': "1", 'techs': "cheap_elec_supply"}].sum().item()
        expensive_prod0 = prod.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].sum().item()
        expensive_prod1 = prod.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].sum().item()
        assert round(cheap_prod1, 5) <= 40
        assert round(expensive_prod0, 5) >= 10
        assert expensive_prod1 == 0

    # All conversion technologies with insufficient supply_max to use the
    # cheap direct heat/cooling supply techs
    def test_carrier_prod_max_non_conversion_all_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_max_non_conversion_all'
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        elec_to_heat = prod.loc[{'locs': "1", 'techs': "elec_to_heat"}].sum().item()
        elec_to_heat_cool_linked = prod.loc[{'locs': "1", 'techs': "elec_to_heat_cool_linked"}].sum('timesteps')
        elec_to_heat_cool_unlinked = prod.loc[{'locs': "1", 'techs': "elec_to_heat_cool_unlinked"}].sum().item()
        cheap_heat = prod.loc[{'locs': "1", 'techs': "cheap_heat_supply"}].sum().item()
        cheap_cool = prod.loc[{'locs': "1", 'techs': "cheap_cool_supply"}].sum().item()
        assert round(cheap_heat, 5) <= 10
        assert round(cheap_cool, 5) <= 5
        assert round(elec_to_heat_cool_linked.loc[{'carriers': 'cool'}].item()) >= 1
        assert round(elec_to_heat_cool_linked.loc[{'carriers': 'heat'}].item()) >= 5
        assert elec_to_heat_cool_unlinked == 0
        assert elec_to_heat == 0

    # All technologies, but insufficient supply_max for enough installed capacity to meet demand
    @pytest.mark.filterwarnings("ignore:(?s).*`['demand', 'storage', 'transmission']` have been ignored*:calliope.exceptions.ModelWarning")
    def test_carrier_prod_max_all_techs_infeasible_constraint(self, model_file):
        model = build_model(
            model_file=model_file,
            scenario='carrier_prod_max_all_techs_infeasible'
        )

        model.run()

        assert model._model_data.attrs['termination_condition'] != 'optimal'



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

examples.py
~~~~~~~~~~~

Example models.

"""

import os

from calliope.core.model import Model


_PATHS = {
    'national_scale': os.path.join(os.path.dirname(__file__), 'example_models', 'national_scale'),
    'urban_scale': os.path.join(os.path.dirname(__file__), 'example_models', 'urban_scale')
}


def national_scale(*args, **kwargs):
    """Returns the built-in national-scale example model."""
    return Model(os.path.join(_PATHS['national_scale'], 'model.yaml'), *args, **kwargs)


def time_clustering(*args, **kwargs):
    """Returns the built-in national-scale example model with time clustering."""
    return Model(
        os.path.join(_PATHS['national_scale'], 'model.yaml'),
        scenario='time_clustering',
        *args, **kwargs
    )


def time_resampling(*args, **kwargs):
    """Returns the built-in national-scale example model with time resampling."""
    return Model(
        os.path.join(_PATHS['national_scale'], 'model.yaml'),
        scenario='time_resampling',
        *args, **kwargs
    )


def urban_scale(*args, **kwargs):
    """Returns the built-in urban-scale example model."""
    return Model(os.path.join(_PATHS['urban_scale'], 'model.yaml'), *args, **kwargs)


def milp(*args, **kwargs):
    """Returns the built-in urban-scale example model with MILP constraints enabled."""
    return Model(
        os.path.join(_PATHS['urban_scale'], 'model.yaml'),
        scenario='milp',
        *args, **kwargs
    )


def operate(*args, **kwargs):
    """Returns the built-in urban-scale example model in operate mode."""
    return Model(
        os.path.join(_PATHS['urban_scale'], 'model.yaml'),
        scenario='operate',
        *args, **kwargs
    )


def time_masking(*args, **kwargs):
    """Returns the built-in urban-scale example model with time masking."""
    return Model(
        os.path.join(_PATHS['urban_scale'], 'model.yaml'),
        scenario='time_masking',
        *args, **kwargs
    )




from calliope.core.preprocess.model_run import \
    model_run_from_yaml, \
    model_run_from_dict, \
    generate_model_run, \
    apply_overrides

from calliope.core.preprocess.model_data import build_model_data

from calliope.core.preprocess.time import \
    apply_time_clustering, \
    final_timedimension_processing



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

conversion.py
~~~~~~~~~~~~~~~~~

Conversion technology constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import \
    get_param

ORDER = 20  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data['sets']

    if 'loc_techs_balance_conversion_constraint' in sets:
        backend_model.balance_conversion_constraint = po.Constraint(
            backend_model.loc_techs_balance_conversion_constraint, backend_model.timesteps,
            rule=balance_conversion_constraint_rule
        )

    if 'loc_techs_cost_var_conversion_constraint' in sets:
        backend_model.cost_var_conversion_constraint = po.Constraint(
            backend_model.costs, backend_model.loc_techs_cost_var_conversion_constraint,
            backend_model.timesteps,
            rule=cost_var_conversion_constraint_rule
        )


def balance_conversion_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance energy carrier consumption and production

    .. container:: scrolling-wrapper

        .. math::

            -1 * \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            \\times \\eta_{energy}(loc::tech, timestep)
            = \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\times \\eta_{energy}(loc::tech, timestep)
            \\quad \\forall loc::tech \\in locs::techs_{conversion},
            \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data['data']

    loc_tech_carrier_out = model_data_dict['lookup_loc_techs_conversion'][('out', loc_tech)]
    loc_tech_carrier_in = model_data_dict['lookup_loc_techs_conversion'][('in', loc_tech)]

    energy_eff = get_param(backend_model, 'energy_eff', (loc_tech, timestep))

    return (
        backend_model.carrier_prod[loc_tech_carrier_out, timestep] == -1 *
        backend_model.carrier_con[loc_tech_carrier_in, timestep] * energy_eff
    )


def cost_var_conversion_constraint_rule(backend_model, cost, loc_tech, timestep):
    """
    Add time-varying conversion technology costs

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{var}}(loc::tech, cost, timestep) =
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\times timestep_{weight}(timestep) \\times cost_{om, prod}(loc::tech, cost, timestep)
            +
            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            \\times timestep_{weight}(timestep) \\times cost_{om, con}(loc::tech, cost, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{cost_{var}, conversion}
    """
    model_data_dict = backend_model.__calliope_model_data
    weight = backend_model.timestep_weights[timestep]

    loc_tech_carrier_in = (
        model_data_dict['data']['lookup_loc_techs_conversion'][('in', loc_tech)]
    )

    loc_tech_carrier_out = (
        model_data_dict['data']['lookup_loc_techs_conversion'][('out', loc_tech)]
    )

    cost_om_prod = get_param(backend_model, 'cost_om_prod',
                                (cost, loc_tech, timestep))
    cost_om_con = get_param(backend_model, 'cost_om_con',
                            (cost, loc_tech, timestep))
    if po.value(cost_om_prod):
        cost_prod = (cost_om_prod * weight *
            backend_model.carrier_prod[loc_tech_carrier_out, timestep])
    else:
        cost_prod = 0

    if po.value(cost_om_con):
        cost_con = (cost_om_con * weight * -1 *
            backend_model.carrier_con[loc_tech_carrier_in, timestep])
    else:
        cost_con = 0

    backend_model.cost_var_rhs[cost, loc_tech, timestep] = cost_prod + cost_con

    return (backend_model.cost_var[cost, loc_tech, timestep] ==
            backend_model.cost_var_rhs[cost, loc_tech, timestep])



import pytest  # pylint: disable=unused-import
import calliope
import logging
import datetime
import os
import tempfile

import xarray as xr
import pandas as pd

from calliope.core.util import dataset, observed_dict

from calliope.core.util.tools import \
    memoize, \
    memoize_instancemethod

from calliope import exceptions
from calliope.core.util.logging import log_time
from calliope.core.util.generate_runs import generate_runs
from calliope.test.common.util import (
    python36_or_higher,
    check_error_or_warning,
    build_test_model
)

_MODEL_NATIONAL = os.path.join(
    os.path.dirname(__file__), '..',
    'example_models', 'national_scale', 'model.yaml'
)

_MODEL_URBAN = os.path.join(
    os.path.dirname(__file__), '..',
    'example_models', 'urban_scale', 'model.yaml'
)


class TestDataset:
    @pytest.fixture()
    def loc_techs(self):
        loc_techs = [
            'region1-3::csp', 'region1::demand_power', 'region1-1::csp',
            'region1-2::csp', 'region2::demand_power', 'region1-1::demand_power'
        ]
        return loc_techs

    @pytest.fixture()
    def example_dataarray(self):
        return xr.DataArray(
            [[[0], [1], [2]], [[3], [4], [5]]], dims=('timesteps', 'loc_techs_bar', 'costs'),
            coords={'timesteps': ['foo', 'bar'], 'loc_techs_bar': ['1::foo', '2::bar', '3::baz'],
                    'costs': ['foo']}
        )

    @pytest.fixture()
    def example_one_dim_dataarray(self):
        return xr.DataArray([0, 1, 2], dims=('timesteps'), coords={'timesteps': ['foo', 'bar', 'baz']})

    @pytest.fixture()
    def example_dataset(self, example_dataarray):
        return xr.Dataset({'foo': example_dataarray, 'bar': example_dataarray.squeeze()})

    def test_get_loc_techs_tech(self, loc_techs):
        loc_techs = dataset.get_loc_techs(loc_techs, tech='csp')
        assert loc_techs == [
            'region1-3::csp', 'region1-1::csp', 'region1-2::csp'
        ]

    def test_get_loc_techs_loc(self, loc_techs):
        loc_techs = dataset.get_loc_techs(loc_techs, loc='region1-1')
        assert loc_techs == [
            'region1-1::csp', 'region1-1::demand_power'
        ]

    def test_get_loc_techs_loc_and_tech(self, loc_techs):
        loc_techs = dataset.get_loc_techs(
            loc_techs, tech='demand_power', loc='region1')
        assert loc_techs == ['region1::demand_power']

    def test_split_loc_tech_to_dataarray(self, example_dataarray):
        formatted_array = dataset.split_loc_techs(example_dataarray)
        assert isinstance(formatted_array, xr.DataArray)
        assert formatted_array.dims == ('costs', 'locs', 'techs', 'timesteps')

    def test_split_loc_tech_to_series(self, example_dataarray):
        formatted_series = dataset.split_loc_techs(example_dataarray, return_as='Series')
        assert isinstance(formatted_series, pd.Series)
        assert formatted_series.index.names == ['costs', 'locs', 'techs', 'timesteps']

    def test_split_loc_tech_to_multiindex_dataarray(self, example_dataarray):
        formatted_array = dataset.split_loc_techs(example_dataarray, return_as='MultiIndex DataArray')
        assert isinstance(formatted_array, xr.DataArray)
        assert formatted_array.dims == ('timesteps', 'loc_techs_bar', 'costs')
        assert isinstance(formatted_array.loc_techs_bar.to_index(), pd.MultiIndex)

    def test_split_loc_tech_too_many_loc_tech_dims(self, example_dataarray):
        _array = example_dataarray.rename({'costs': 'loc_techs_2'})
        with pytest.raises(exceptions.ModelError) as excinfo:
            dataset.split_loc_techs(_array)
        assert check_error_or_warning(
            excinfo, 'Cannot split loc_techs or loc_tech_carriers dimension'
        )

    def test_split_loc_tech_one_dim_to_dataarray(self, example_one_dim_dataarray):
        formatted_array = dataset.split_loc_techs(example_one_dim_dataarray)
        assert isinstance(formatted_array, xr.DataArray)
        assert formatted_array.dims == ('timesteps',)

    def test_split_loc_tech_one_dim_to_series(self, example_one_dim_dataarray):
        formatted_series = dataset.split_loc_techs(example_one_dim_dataarray, return_as='Series')
        assert isinstance(formatted_series, pd.Series)
        assert formatted_series.index.names == ['timesteps']

    def test_split_loc_tech_one_dim_to_multiindex_dataarray(self, example_one_dim_dataarray):
        formatted_array = dataset.split_loc_techs(example_one_dim_dataarray, return_as='MultiIndex DataArray')
        assert isinstance(formatted_array, xr.DataArray)
        assert formatted_array.dims == ('timesteps',)

    def test_split_loc_tech_unknown_output(self, example_dataarray, example_one_dim_dataarray):
        for array in [example_dataarray, example_one_dim_dataarray]:
            with pytest.raises(ValueError) as excinfo:
                dataset.split_loc_techs(array, return_as='foo')
            assert check_error_or_warning(
                excinfo,
                '`return_as` must be `DataArray`, `Series`, or `MultiIndex DataArray`'
            )

    def test_reorganise_dataset_dimensions(self, example_dataset):
        reorganised_dataset = dataset.reorganise_xarray_dimensions(example_dataset)
        dataset_dims = [i for i in reorganised_dataset.dims.keys()]
        assert dataset_dims == ['costs', 'loc_techs_bar', 'timesteps']

    def test_reorganise_dataarray_dimensions(self, example_dataarray):
        reorganised_dataset = dataset.reorganise_xarray_dimensions(example_dataarray)
        assert reorganised_dataset.dims == ('costs', 'loc_techs_bar', 'timesteps')

    def test_fail_reorganise_dimensions(self):
        with pytest.raises(TypeError) as excinfo:
            dataset.reorganise_xarray_dimensions(['timesteps', 'loc_techs_bar', 'costs'])
        assert check_error_or_warning(
            excinfo, 'Must provide either xarray Dataset or DataArray to be reorganised'
        )


class TestMemoization:
    @memoize_instancemethod
    def instance_method(self, a, b):
        return a + b

    def test_memoize_one_arg(self):
        @memoize
        def test(a):
            return a + 1
        assert test(1) == 2
        assert test(1) == 2

    def test_memoize_two_args(self):
        @memoize
        def test(a, b):
            return a + b
        assert test(1, 2) == 3
        assert test(1, 2) == 3

    def test_memoize_instancemethod(self):
        assert self.instance_method(1, 2) == 3
        assert self.instance_method(1, 2) == 3


class TestLogging:
    def test_set_log_verbosity(self):
        calliope.set_log_verbosity('CRITICAL', include_solver_output=True)

        assert logging.getLogger('calliope').getEffectiveLevel() == 50
        assert logging.getLogger('py.warnings').getEffectiveLevel() == 50
        assert logging.getLogger('calliope.backend.pyomo.model').getEffectiveLevel() == 10

        calliope.set_log_verbosity('CRITICAL', include_solver_output=False)

        assert logging.getLogger('calliope').getEffectiveLevel() == 50
        assert logging.getLogger('py.warnings').getEffectiveLevel() == 50
        assert logging.getLogger('calliope.backend.pyomo.model').getEffectiveLevel() == 50

        calliope.set_log_verbosity()

        assert logging.getLogger('calliope').getEffectiveLevel() == 20
        assert logging.getLogger('py.warnings').getEffectiveLevel() == 20
        assert logging.getLogger('calliope.backend.pyomo.model').getEffectiveLevel() == 10

    def test_timing_log(self):
        timings = {'model_creation': datetime.datetime.now()}
        logger = logging.getLogger('calliope.testlogger')

        # TODO: capture logging output and check that comment is in string
        log_time(logger, timings, 'test', comment='test_comment', level='info')
        assert isinstance(timings['test'], datetime.datetime)

        log_time(logger, timings, 'test2', comment=None, level='info')
        assert isinstance(timings['test2'], datetime.datetime)

        # TODO: capture logging output and check that time_since_run_start is in the string
        log_time(logger, timings, 'test', comment=None, level='info', time_since_run_start=True)


class TestGenerateRuns:

    @python36_or_higher
    def test_generate_runs_scenarios(self):
        runs = generate_runs(
            _MODEL_NATIONAL,
            scenarios='time_resampling;profiling;time_clustering'
        )
        assert len(runs) == 3
        assert runs[0].endswith(
            '--scenario time_resampling --save_netcdf out_1_time_resampling.nc --save_plots plots_1_time_resampling.html'
        )

    @python36_or_higher
    def test_generate_runs_scenarios_none_with_scenarios(self):
        runs = generate_runs(
            _MODEL_NATIONAL,
            scenarios=None
        )
        assert len(runs) == 2
        assert runs[0].endswith(
            '--scenario cold_fusion_with_production_share --save_netcdf out_1_cold_fusion_with_production_share.nc --save_plots plots_1_cold_fusion_with_production_share.html'
        )

    @python36_or_higher
    def test_generate_runs_scenarios_none_with_overrides(self):
        runs = generate_runs(
            _MODEL_URBAN,
            scenarios=None,
        )
        assert len(runs) == 4
        assert runs[0].endswith(
            '--scenario milp --save_netcdf out_1_milp.nc --save_plots plots_1_milp.html'
        )


class TestPandasExport:

    @pytest.fixture(scope="module")
    def model(self):
        return calliope.examples.national_scale()

    @pytest.mark.parametrize(
        "variable_name",
        sorted([i for i in calliope.examples.national_scale()._model_data.data_vars.keys()])
    )
    def test_data_variables_can_be_exported_to_pandas(self, model, variable_name):
        model.get_formatted_array(variable_name).to_dataframe()


class TestObservedDict:

    def as_yaml(self, _dict, strip=False):
        if strip is True:
            _dict = {
                k: v for k, v in _dict.items()
                if (not isinstance(v, dict) and v is not None)
                or (isinstance(v, dict) and len(v.keys()) > 0)
            }
        return calliope.AttrDict.to_yaml(calliope.AttrDict(_dict))

    @pytest.fixture(scope="module")
    def model(self):
        return calliope.examples.national_scale()

    @pytest.fixture(scope="module")
    def observer(self):
        return xr.Dataset()

    @pytest.fixture(scope="module")
    def observed_from_dict(self, observer):
        initial_dict = {'foo': 'bar', 'foobar': {'baz': 'fob'}}
        return observed_dict.UpdateObserverDict(
            initial_dict=initial_dict, name='test', observer=observer
        )

    @pytest.fixture(scope="module")
    def observed_from_string(self, observer):
        initial_dict = {'foo': 'bar', 'foobar': {'baz': 'fob'}}
        return observed_dict.UpdateObserverDict(
            initial_yaml_string=self.as_yaml(initial_dict), name='test_2', observer=observer
        )

    def test_initialise_observer(self, observer, observed_from_dict, observed_from_string):
        assert 'test' in observer.attrs.keys()
        assert 'test_2' in observer.attrs.keys()
        assert observer.attrs['test'] == self.as_yaml({'foo': 'bar', 'foobar': {'baz': 'fob'}})
        observer.attrs['test'] == observer.attrs['test_2']

    def test_value_error_on_initialising(self):
        initial_dict = {'foo': 'bar', 'foobar': {'baz': 'fob'}}
        initial_string = self.as_yaml(initial_dict)
        with pytest.raises(ValueError) as error:
            observed_dict.UpdateObserverDict(
                name='test_2', observer=xr.Dataset()
            )
        assert check_error_or_warning(
            error,
            'must supply one, and only one, of initial_dict or initial_yaml_string'
        )
        with pytest.raises(ValueError) as error:
            observed_dict.UpdateObserverDict(
                initial_yaml_string=initial_string,
                initial_dict=initial_dict,
                name='test_2', observer=xr.Dataset()
            )
        assert check_error_or_warning(
            error,
            'must supply one, and only one, of initial_dict or initial_yaml_string'
        )

    @pytest.mark.parametrize('key1,key2,value,result', (
        ('foo', None, 1, {'foo': 1, 'foobar': {'baz': 'fob'}}),
        ('foobar', 'baz', 2, {'foo': 1, 'foobar': {'baz': 2}}),
        ('foo', None, {'baz': 'fob'}, {'foo': {'baz': 'fob'}, 'foobar': {'baz': 2}}),
        ('foo', 'baz', 3, {'foo': {'baz': 3}, 'foobar': {'baz': 2}}),
        ('foo', None, {}, {'foobar': {'baz': 2}}),
        ('foo', None, 5, {'foo': 5, 'foobar': {'baz': 2}}),
        ('foo', None, None, {'foobar': {'baz': 2}})

    ))
    def test_set_item_observer(self, observed_from_dict, observer, key1, key2, value, result):
        if key2 is None:
            observed_from_dict[key1] = value
        else:
            observed_from_dict[key1][key2] = value

        assert observer.attrs['test'] == self.as_yaml(result)

    def test_update_observer(self, observed_from_dict, observer):
        observed_from_dict.update({'baz': 4})
        assert observer.attrs['test'] == self.as_yaml({'foobar': {'baz': 2}, 'baz': 4})

    def test_reinstate_observer(self, observed_from_dict, observer):
        observer.attrs['test'] = "{}"
        assert observer.attrs['test'] == "{}"
        observed_from_dict['foo'] = 5
        assert observer.attrs['test'] == self.as_yaml({'foo': 5, 'foobar': {'baz': 2}, 'baz': 4})

    def test_model_config(self, model):
        assert hasattr(model, 'model_config')
        assert 'model_config' in model._model_data.attrs.keys()
        assert model._model_data.attrs['model_config'] == self.as_yaml(model.model_config, strip=True)

        model.model_config['name'] = 'new name'
        assert model.model_config['name'] == 'new name'
        assert model._model_data.attrs['model_config'] == self.as_yaml(model.model_config, strip=True)

    def test_run_config(self, model):
        assert hasattr(model, 'run_config')
        assert 'run_config' in model._model_data.attrs.keys()
        assert model._model_data.attrs['run_config'] == self.as_yaml(model.run_config, strip=True)

        model.run_config['solver'] = 'cplex'
        assert model.run_config['solver'] == 'cplex'
        assert model._model_data.attrs['run_config'] == self.as_yaml(model.run_config, strip=True)

    def test_load_from_netcdf(self, model):

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'model.nc')
            model.to_netcdf(out_path)

            model_from_disk = calliope.read_netcdf(out_path)
            assert hasattr(model_from_disk, 'run_config')
            assert 'run_config' in model_from_disk._model_data.attrs.keys()
            assert hasattr(model_from_disk, 'model_config')
            assert 'model_config' in model_from_disk._model_data.attrs.keys()



import os
import tempfile

import pytest  # pylint: disable=unused-import
from click.testing import CliRunner

import calliope
from calliope import cli, AttrDict

_THIS_DIR = os.path.dirname(__file__)
_MODEL_NATIONAL = os.path.join(
    _THIS_DIR,
    '..', 'example_models', 'national_scale', 'model.yaml'
)
_MINIMAL_TEST_MODEL = os.path.join(
    _THIS_DIR, 'common', 'test_model', 'model_minimal.yaml'
)


class TestCLI:
    def test_new(self):
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tempdir:
            new_path = os.path.join(tempdir, 'test')
            result = runner.invoke(cli.new, [new_path])
            assert result.exit_code == 0
            # Assert that `model.yaml` in the target dir exists
            assert os.path.isfile(os.path.join(tempdir, 'test', 'model.yaml'))

    def test_run_from_yaml(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.run, [_MODEL_NATIONAL, '--save_netcdf=output.nc', '--save_plots=results.html'])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, 'output.nc'))
            assert os.path.isfile(os.path.join(tempdir, 'results.html'))

    def test_incorrect_file_format(self):
        runner = CliRunner()

        result = runner.invoke(cli.run, ['test_model.txt', '--save_netcdf=output.nc'])
        assert 'Cannot determine model file format' in result.output
        assert result.exit_code == 1

    def test_incorrect_model_format(self):
        runner = CliRunner()

        result = runner.invoke(cli.run, ['test_model.txt', '--model_format=yml', '--save_netcdf=output.nc'])
        assert 'Invalid model format' in result.output
        assert result.exit_code == 1

    @pytest.mark.parametrize('arg', (('--scenario=test'), ("--override_dict={'model.name': 'test'}")))
    def test_unavailable_arguments(self, arg):
        runner = CliRunner()

        result = runner.invoke(cli.run, ['test_model.nc', arg, '--save_netcdf=output.nc'])
        assert 'the --scenario and --override_dict options are not available' in result.output
        assert result.exit_code == 1

    def test_run_from_netcdf(self):
        runner = CliRunner()
        model = calliope.examples.national_scale()
        with runner.isolated_filesystem() as tempdir:
            model_file = os.path.join(tempdir, 'model.nc')
            out_file = os.path.join(tempdir, 'output.nc')
            model.to_netcdf(model_file)
            result = runner.invoke(cli.run, [model_file, '--save_netcdf=output.nc'])
            assert result.exit_code == 0
            assert os.path.isfile(out_file)

    def test_run_save_lp(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.run, [_MODEL_NATIONAL, '--save_lp=output.lp'])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, 'output.lp'))

    def test_generate_runs_bash(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.generate_runs, [
                _MODEL_NATIONAL, 'test.sh', '--kind=bash',
                '--scenarios="run1;run2;run3;run4"'
            ])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, 'test.sh'))

    def test_generate_runs_windows(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.generate_runs, [
                _MODEL_NATIONAL, 'test.bat', '--kind=windows',
                '--scenarios="run1;run2;run3;run4"'
            ])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, 'test.bat'))

    def test_generate_runs_bsub(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.generate_runs, [
                _MODEL_NATIONAL, 'test.sh', '--kind=bsub',
                '--scenarios="run1;run2;run3;run4"',
                '--cluster_mem=1G', '--cluster_time=100'
            ])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, 'test.sh'))
            assert os.path.isfile(os.path.join(tempdir, 'test.sh.array.sh'))

    def test_generate_runs_sbatch(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.generate_runs, [
                _MODEL_NATIONAL, 'test.sh', '--kind=sbatch',
                '--scenarios="run1;run2;run3;run4"',
                '--cluster_mem=1G', '--cluster_time=100'
            ])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, 'test.sh'))
            assert os.path.isfile(os.path.join(tempdir, 'test.sh.array.sh'))

    def test_debug(self):
        runner = CliRunner()
        result = runner.invoke(cli.run, ['foo.yaml', '--debug'])
        assert result.exit_code == 1
        assert 'Traceback (most recent call last)' in result.output

        result = runner.invoke(cli.run, ['foo.yaml'])
        assert result.exit_code == 1
        assert 'Traceback (most recent call last)' not in result.output

    def test_generate_scenarios(self):
        runner = CliRunner()
        with runner.isolated_filesystem() as tempdir:
            out_file = os.path.join(tempdir, 'scenarios.yaml')
            result = runner.invoke(cli.generate_scenarios, [
                _MODEL_NATIONAL, out_file,
                'cold_fusion',
                'run1;run2',
                'cold_fusion_cap_share;cold_fusion_prod_share',
            ])
            assert result.exit_code == 0
            assert os.path.isfile(out_file)
            scenarios = AttrDict.from_yaml(out_file)
            assert 'scenario_0' not in scenarios['scenarios']
            assert scenarios['scenarios']['scenario_1'] == [
                'cold_fusion', 'run1', 'cold_fusion_cap_share'
            ]

    def test_no_success_exit_code_when_infeasible(self):
        runner = CliRunner()
        result = runner.invoke(cli.run, [
            _MINIMAL_TEST_MODEL,
            "--scenario=investment_costs",  # without these, the model cannot run
            "--override_dict={techs.test_supply_elec.constraints.energy_cap_max: 1}"  # elec supply too low
        ])
        assert result.exit_code != 0

    def test_success_exit_code_when_infeasible_and_demanded(self):
        runner = CliRunner()
        result = runner.invoke(cli.run, [
            _MINIMAL_TEST_MODEL,
            "--no_fail_when_infeasible",
            "--scenario=investment_costs",  # without these, the model cannot run
            "--override_dict={techs.test_supply_elec.constraints.energy_cap_max: 1}"  # elec supply too low
        ])
        assert result.exit_code == 0



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

generate_runs.py
~~~~~~~~~~~~~~~~

Generate scripts to run multiple versions of the same model
in parallel on a cluster or sequentially on any machine.

"""

import os

import pandas as pd
from calliope.core import AttrDict


def generate_runs(
        model_file, scenarios=None,
        additional_args=None, override_dict=None):
    """
    Returns a list of "calliope run" invocations.

    ``scenarios`` must be specified as either a semicolon-separated
    list of scenarios or a semicolon-separated list of comma-separated
    individual override combinations, such as:

        ``override1,override2;override1,override3;...``

    If ``scenarios`` is not given, use all scenarios in the model
    configuration, and if no scenarios given in the model configuration,
    uses all individual overrides, one by one.

    """
    if scenarios is None:
        config = AttrDict.from_yaml(model_file)
        if override_dict:
            override = AttrDict.from_yaml_string(override_dict)
            config.union(
                override, allow_override=True, allow_replacement=True
            )

        if 'scenarios' in config:
            runs = config.scenarios.keys()
        else:
            runs = config.overrides.keys()
    else:
        runs = scenarios.split(';')

    commands = []

    # len(str(x)) gives us the number of digits in x, for padding
    i_string = '{:0>' + str(len(str(len(runs)))) + 'd}'

    for i, run in enumerate(runs):
        cmd = (
            'calliope run {model} --scenario {scenario} '
            '--save_netcdf out_{i}_{scenario}.nc '
            '--save_plots plots_{i}_{scenario}.html'
        ).format(
            i=i_string.format(i + 1),
            model=model_file,
            scenario=run,
            override_dict=override_dict,
        ).strip()

        if override_dict:
            cmd = cmd + ' --override_dict="{}"'.format(override_dict)

        if additional_args:
            cmd = cmd + ' ' + additional_args

        commands.append(cmd)

    return commands


def generate_bash_script(
        out_file, model_file, scenarios,
        additional_args=None, override_dict=None, **kwargs):

    commands = generate_runs(model_file, scenarios, additional_args, override_dict)

    base_string = '    {i}) {cmd} ;;\n'
    lines_start = [
        '#!/bin/sh',
        '',
        'function process_case () {',
        '    case "$1" in',
        ''
    ]
    lines_end = [
        '    esac', '}',
        '',
        'if [[ $# -eq 0 ]] ; then',
        '    echo No parameter given, running all runs sequentially...',
        '    for i in $(seq 1 {}); do process_case $i; done'.format(len(commands)),
        'else',
        '    echo Running run $1',
        '    process_case $1',
        'fi',
        '',
    ]

    lines_all = lines_start + [base_string.format(i=i + 1, cmd=cmd) for i, cmd in enumerate(commands)] + lines_end

    with open(out_file, 'wb') as f:
        f.write(bytes('\n'.join(lines_all), 'UTF-8'))

    os.chmod(out_file, 0o755)

    return commands


def generate_bsub_script(out_file, model_file, scenarios,
                         additional_args, override_dict,
                         cluster_mem, cluster_time, cluster_threads=1,
                         **kwargs):

    # We also need to generate the bash script to run on the cluster
    bash_out_file = out_file + '.array.sh'
    bash_out_file_basename = os.path.basename(bash_out_file)
    commands = generate_bash_script(
        bash_out_file, model_file, scenarios, additional_args, override_dict)

    lines = [
        '#!/bin/sh',
        '#BSUB -J calliope[1-{}]'.format(len(commands)),
        '#BSUB -n {}'.format(cluster_threads),
        '#BSUB -R "rusage[mem={}]"'.format(cluster_mem),
        '#BSUB -W {}'.format(cluster_time),
        '#BSUB -r',  # Automatically restart failed jobs
        '#BSUB -o log_%I.log',
        '',
        './' + bash_out_file_basename + ' ${LSB_JOBINDEX}',
        ''
    ]

    with open(out_file, 'wb') as f:
        f.write(bytes('\n'.join(lines), 'UTF-8'))


def generate_sbatch_script(out_file, model_file, scenarios,
                           additional_args, override_dict,
                           cluster_mem, cluster_time, cluster_threads=1,
                           **kwargs):
    """
    SBATCH (SLURM) script generator.
    """

    # We also need to generate the bash script to run on the cluster
    bash_out_file = out_file + '.array.sh'
    bash_out_file_basename = os.path.basename(bash_out_file)
    commands = generate_bash_script(
        bash_out_file, model_file, scenarios, additional_args, override_dict)

    if ':' not in cluster_time:
        # Assuming time given as minutes, so needs changing to %H:%M%S
        cluster_time = pd.to_datetime(cluster_time, unit='m').strftime('%H:%M:%S')

    lines = [
        '#!/bin/bash',
        '#SBATCH -J calliope',  # Name of the job
        '#SBATCH --array=1-{}'.format(len(commands)),  # How many jobs there are
        '#SBATCH --ntasks={}'.format(cluster_threads),
        '#SBATCH --mem={}'.format(cluster_mem),
        '#SBATCH --time={}'.format(cluster_time),  # How much wallclock time will be required
        '#SBATCH -o log_%a.log',
        '',
        '#! Optional add-ins for SBATCH (uncomment and add info as necessary):',
        '##SBATCH -A project_name',  # Which project should be charged'
        '##SBATCH --nodes=X',  # X whole nodes should be allocated'
        '##SBATCH -p partition_name',
        '',
        '#! Insert module load commands after this line, if needed:',
        '#! (Note: you can load this in ~.bashrc if you want them loaded every time you log in)',
        '',
        '#! module load gurobi (or glpk/cplex)',
        '#! module load miniconda3',
        '#! module load /path/to/miniconda3/envs/your_env_name/',
        '',
        'cd $SLURM_SUBMIT_DIR',
        '',
        './' + bash_out_file_basename + ' ${SLURM_ARRAY_TASK_ID}'
    ]

    with open(out_file, 'wb') as f:
        f.write(bytes('\n'.join(lines), 'UTF-8'))


def generate_windows_script(
        out_file, model_file, scenarios,
        additional_args=None, override_dict=None,
        **kwargs):

    commands = generate_runs(
        model_file, scenarios, additional_args, override_dict)

    # \r\n are Windows line endings
    base_string = 'echo "Run {i}"\r\n{cmd}\r\n'
    lines_start = [
        '@echo off',
        '',
    ]

    lines_all = lines_start + [
        base_string.format(i=i + 1, cmd=cmd)
        for i, cmd in enumerate(commands)]

    with open(out_file, 'wb') as f:
        f.write(bytes('\r\n'.join(lines_all), 'UTF-8'))

    os.chmod(out_file, 0o755)

    return commands


_KINDS = {
    'bash': generate_bash_script,
    'bsub': generate_bsub_script,
    'windows': generate_windows_script,
    'sbatch': generate_sbatch_script
}


def generate(kind, **kwargs):
    _KINDS[kind](**kwargs)



import pytest  # pylint: disable=unused-import

from calliope.test.common.util import build_test_model as build_model
from calliope.test.common.util import check_variable_exists


class TestBuildConversionPlusConstraints:
    # conversion_plus.py
    def test_loc_techs_balance_conversion_plus_primary_constraint(self):
        """
        sets.loc_techs_conversion_plus,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_plus_primary_constraint')

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_primary_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials.carrier_out': ['electricity', 'heat']},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_primary_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in': ['coal', 'gas'], 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_primary_constraint')

    def test_loc_techs_carrier_production_max_conversion_plus_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if i not in sets.loc_techs_milp
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_production_max_conversion_plus_constraint')

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_techs_carrier_production_max_conversion_plus_milp_constraint(self):
        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_max_conversion_plus_constraint')

    def test_loc_techs_carrier_production_min_conversion_plus_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i not in sets.loc_techs_milp
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_conversion_plus_constraint')

        m = build_model(
            {'techs.test_conversion_plus.constraints.energy_cap_min_use': 0.1},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_production_min_conversion_plus_constraint')

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_techs_carrier_production_min_conversion_plus_milp_constraint(self):
        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_conversion_plus_constraint')

        m = build_model(
            {'techs.test_conversion_plus.constraints.energy_cap_min_use': 0.1},
            'conversion_plus_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_conversion_plus_constraint')

    def test_loc_techs_cost_var_conversion_plus_constraint(self):
        """
        sets.loc_techs_om_cost_conversion_plus,
        """
        # no conversion_plus = no constraint
        m = build_model(
            {'techs.test_supply_elec.costs.monetary.om_prod': 0.1},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')

        # no conversion_plus = no constraint
        m = build_model(
            {'techs.test_conversion.costs.monetary.om_prod': 0.1},
            'simple_conversion,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')

        # no variable costs for conversion_plus = no constraint
        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')

        # om_prod creates constraint and populates it with carrier_prod driven cost
        m = build_model(
            {'techs.test_conversion_plus.costs.monetary.om_prod': 0.1},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')
        assert check_variable_exists(
            m._backend_model, 'cost_var_conversion_plus_constraint', 'carrier_prod'
        )
        assert not check_variable_exists(
            m._backend_model, 'cost_var_conversion_plus_constraint', 'carrier_con'
        )

        # om_con creates constraint and populates it with carrier_con driven cost
        m = build_model(
            {'techs.test_conversion_plus.costs.monetary.om_con': 0.1},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')
        assert check_variable_exists(
            m._backend_model, 'cost_var_conversion_plus_constraint', 'carrier_con'
        )
        assert not check_variable_exists(
            m._backend_model, 'cost_var_conversion_plus_constraint', 'carrier_prod'
        )

    def test_loc_techs_balance_conversion_plus_in_2_constraint(self):
        """
        sets.loc_techs_in_2,
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_plus_in_2_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in_2': 'coal', 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_in_2_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in_2': ['coal', 'heat'], 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_in_2_constraint')

    def test_loc_techs_balance_conversion_plus_in_3_constraint(self):
        """
        sets.loc_techs_in_3,
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_plus_in_3_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in_3': 'coal', 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_in_3_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in_3': ['coal', 'heat'], 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_in_3_constraint')

    def test_loc_techs_balance_conversion_plus_out_2_constraint(self):
        """
        sets.loc_techs_out_2,
        """

        m = build_model(
            {'techs.test_conversion_plus.essentials.carrier_out_2': ['coal', 'heat']},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_out_2_constraint')

    def test_loc_techs_balance_conversion_plus_out_3_constraint(self):
        """
        sets.loc_techs_out_3,
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_plus_out_3_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials.carrier_out_3': 'coal'},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_out_3_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials.carrier_out_3': ['coal', 'heat']},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_out_3_constraint')

    def test_loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint(self):
        """
        """

        m = build_model({}, 'simple_conversion_plus,one_day,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'conversion_plus_prod_con_to_zero_constraint')

        m = build_model(
            {'techs.test_conversion_plus.constraints.carrier_ratios.carrier_out_2.heat': 'file=carrier_ratio.csv'},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'conversion_plus_prod_con_to_zero_constraint')

        m = build_model(
            {'techs.test_conversion_plus.constraints.carrier_ratios.carrier_out_2.heat': 'file=carrier_ratio.csv'},
            'simple_conversion_plus,one_day,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'conversion_plus_prod_con_to_zero_constraint')
        assert check_variable_exists(
            m._backend_model,
            'conversion_plus_prod_con_to_zero_constraint',
            'carrier_prod'
        )
        assert hasattr(m._backend_model, 'loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint')

        m = build_model(
            {'techs.test_conversion_plus': {
                'essentials.carrier_in': ['gas', 'coal'],
                'essentials.primary_carrier_in': 'gas',
                'constraints.carrier_ratios': {
                    'carrier_out_2.heat': 'file=carrier_ratio.csv',
                    'carrier_in.coal': 'file=carrier_ratio.csv'
                }
            }},
            'simple_conversion_plus,one_day,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'conversion_plus_prod_con_to_zero_constraint')
        assert check_variable_exists(
            m._backend_model,
            'conversion_plus_prod_con_to_zero_constraint',
            'carrier_prod'
        )
        assert check_variable_exists(
            m._backend_model,
            'conversion_plus_prod_con_to_zero_constraint',
            'carrier_con'
        )
        assert hasattr(m._backend_model, 'loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint')


class TestConversionPlusConstraintResults:

    def test_carrier_ratio_from_file(self):
        m = build_model(
            {'techs.test_conversion.costs.monetary.om_prod': 3},
            'conversion_and_conversion_plus,one_day,investment_costs'
        )
        m.run()
        carrier_prod_conversion_plus = (
            m.get_formatted_array('carrier_prod').loc[{'techs': 'test_conversion_plus'}]
        ).sum('locs').to_pandas()
        assert all(
            (carrier_prod_conversion_plus.loc['heat'] /
             carrier_prod_conversion_plus.loc['electricity']).dropna()
            == 0.8
        )

        m = build_model(
            {'techs.test_conversion_plus': {
                'constraints.carrier_ratios': {
                    'carrier_out_2.heat': 'file=carrier_ratio.csv', 'carrier_in.coal': 'file=carrier_ratio.csv'
                },
                'essentials': {'carrier_in': ['gas', 'coal'], 'primary_carrier_in': 'gas'},
                'costs.monetary': {'om_con': 3, 'om_prod': 3}
            }},
            'conversion_and_conversion_plus,one_day,investment_costs'
        )
        m.run()
        carrier_prod_conversion_plus = (
            m.get_formatted_array('carrier_prod').loc[{'techs': 'test_conversion_plus'}]
        ).sum('locs').to_pandas()
        carrier_con_conversion_plus = (
            m.get_formatted_array('carrier_con').loc[{'techs': 'test_conversion_plus'}]
        ).sum('locs').to_pandas()

        prod_ratios = (
            carrier_prod_conversion_plus.loc['heat'] / carrier_prod_conversion_plus.loc['electricity']
        ).dropna()
        con_ratios = (
            carrier_con_conversion_plus.loc['coal'] / carrier_prod_conversion_plus.sum(axis=0)
        ).dropna()

        assert (
            m._model_run.timeseries_data['carrier_ratio.csv'].loc[:, '0'].reindex(prod_ratios.index)
            == prod_ratios.round(1)
        ).all()
        assert (
            m._model_run.timeseries_data['carrier_ratio.csv'].loc[:, '0'].reindex(con_ratios.index)
            == con_ratios.round(1)
        ).all()



import pytest

import calliope
from calliope.test.common.util import build_test_model as build_model


class TestFutureWarnings:

    def test_group_share_warning(self):
        override = {
            'model.group_share.test_supply_elec.carrier_prod_min.electricity': 0.85
        }
        expected_warning_msg = '`group_share` constraints will be removed in v0.7.0'
        with pytest.warns(FutureWarning, match=expected_warning_msg):
            build_model(override_dict=override, scenario='simple_supply,one_day')

    @pytest.mark.filterwarnings('always::FutureWarning')
    def test_default_cost_class_warning(self):
        expected_warning_msg = 'There will be no default cost class for the objective function in v0.7.0'
        with pytest.warns(FutureWarning, match=expected_warning_msg):
            build_model(scenario='simple_supply,one_day')

    @pytest.mark.filterwarnings('error:.*There will be no default cost class for the objective function in v0.7.0.*:FutureWarning')
    def test_default_cost_class_from_override_no_warning(self):
        override = {'run.objective_options.cost_class': {'monetary': 1}}
        build_model(override_dict=override, scenario='simple_supply,one_day')

    @pytest.mark.filterwarnings('error:.*There will be no default cost class for the objective function in v0.7.0.*:FutureWarning')
    def test_default_cost_class_from_scenario_no_warning(self):
        build_model(scenario='simple_supply,one_day,explicit_cost_class')

    @pytest.mark.filterwarnings('always::FutureWarning')
    def test_future_warning_for_charge_rate(self):
        with pytest.warns(FutureWarning, match='`charge_rate` is renamed to `energy_cap_per_storage_cap_max`'):
            calliope.examples.national_scale(
                override_dict={'techs.battery.constraints.charge_rate': 5}
            )



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

import numpy as np

from calliope.core.attrdict import AttrDict


def concat_iterable(iterable, concatenators):
    """
    Take an iterable containing iterables of strings,
    return a list of strings concatenating inner iterables with '::'.

    E.g.:
    ``
    result = concat_iterable([('x', 'y', 'z'), ('1', '2', '3')], [':', ':'])
    result == ['x:y:z', '1:2:3']

    result = concat_iterable([('x', 'y', 'z'), ('1', '2', '3')], ['::', ':'])
    result == ['x::y:z', '1::2:3']
    ``

    """
    if len(iterable) == 0:
        return []
    concats = concatenators + ['']
    string_iter_len = len(iterable[0])
    assert all(len(i) == string_iter_len for i in iterable)
    assert string_iter_len == len(concats)

    return [
        ''.join([string_iter[i] + concats[i] for i in range(string_iter_len)])
        for string_iter in iterable
    ]


def constraint_exists(model_run, loc_tech, search_entry):
    """
    Check if a cost/constraint exists for a loc_tech.

    E.g.:
    ``
    constraint_exists(model_run, 'X1::chp', 'constraints.energy_eff') will return True or False
    ``
    """
    if loc_tech in model_run.sets.loc_techs_transmission:
        loc_tech_dict = split_loc_techs_transmission(loc_tech)
        search_string = (
            'locations.{loc_from}.links.{loc_to}.techs.{tech}.{}'
            .format(search_entry, **loc_tech_dict)
        )
    else:
        search_string = (
            'locations.{0}.techs.{1}.{2}'
            .format(*loc_tech.split('::'), search_entry)
        )
    return model_run.get_key(search_string, None)


def get_all_carriers(config, direction='both'):
    if direction == 'both':
        carrier_list = ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3']
    elif direction == 'in':
        carrier_list = ['in', 'in_2', 'in_3']
    elif direction == 'out':
        carrier_list = ['out', 'out_2', 'out_3']

    carriers = flatten_list(
        [config.get_key('carrier', '')] + [
            config.get_key('carrier_{}'.format(k), '')
            for k in carrier_list
        ]
    )

    return set(carriers) - set([''])


def flatten_list(unflattened_list):
    """
    Take list of iterables/non-iterables and outputs a list of non-iterables.
    """
    flattened_list = []
    for item in unflattened_list:
        if hasattr(item, '__iter__') and not isinstance(item, str):
            flattened_list.extend(item)
        else:
            flattened_list.append(item)

    return flattened_list


def split_loc_techs_transmission(transmission_string):
    """
    from loc::tech:link get out a dictionary of {loc_from:loc, loc_to:link, tech:tech}
    """
    loc, tech_link = transmission_string.split('::')
    tech, link = tech_link.split(':')

    return {'loc_from': loc, 'loc_to': link, 'tech': tech}


def get_systemwide_constraints(tech_config):
    if 'constraints' in tech_config:
        constraints = AttrDict({
            k: tech_config.constraints[k]
            for k in tech_config.constraints.keys()
            if k.endswith('_systemwide')
        })
    else:
        constraints = AttrDict({})

    return constraints


def vincenty(coord1, coord2):
    """
    Vincenty's inverse method formula to calculate the distance in metres
    between two points on the surface of a spheroid (WGS84).
    modified from https://github.com/maurycyp/vincenty
    """

    a = 6378137  # equitorial radius in meters
    f = 1 / 298.257223563  # flattening from sphere to oblate spheroid
    b = a * (1 - f)  # polar radius in meters

    max_iter = 200
    thresh = 1e-12

    # short-circuit coincident points
    if coord1[0] == coord2[0] and coord1[1] == coord2[1]:
        return 0

    U1 = np.arctan((1 - f) * np.tan(np.radians(coord1[0])))
    U2 = np.arctan((1 - f) * np.tan(np.radians(coord2[0])))
    L = np.radians(coord2[1] - coord1[1])
    Lambda = L

    sinU1 = np.sin(U1)
    cosU1 = np.cos(U1)
    sinU2 = np.sin(U2)
    cosU2 = np.cos(U2)

    for iteration in range(max_iter):
        sinLambda = np.sin(Lambda)
        cosLambda = np.cos(Lambda)
        sinSigma = np.sqrt((cosU2 * sinLambda) ** 2 +
                             (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) ** 2)
        if sinSigma == 0:
            return 0.0  # coincident points
        cosSigma = sinU1 * sinU2 + cosU1 * cosU2 * cosLambda
        sigma = np.arctan2(sinSigma, cosSigma)
        sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
        cosSqAlpha = 1 - sinAlpha ** 2
        try:
            cos2SigmaM = cosSigma - 2 * sinU1 * sinU2 / cosSqAlpha
        except ZeroDivisionError:
            cos2SigmaM = 0
        C = f / 16 * cosSqAlpha * (4 + f * (4 - 3 * cosSqAlpha))
        LambdaPrev = Lambda
        Lambda = L + (1 - C) * f * sinAlpha * (sigma + C * sinSigma *
                                               (cos2SigmaM + C * cosSigma *
                                                (-1 + 2 * cos2SigmaM ** 2)))
        if abs(Lambda - LambdaPrev) < thresh:
            break  # successful convergence
    else:
        return None  # failure to converge

    uSq = cosSqAlpha * (a ** 2 - b ** 2) / (b ** 2)
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))
    deltaSigma = B * sinSigma * (cos2SigmaM + B / 4 * (cosSigma *
                 (-1 + 2 * cos2SigmaM ** 2) - B / 6 * cos2SigmaM *
                 (-3 + 4 * sinSigma ** 2) * (-3 + 4 * cos2SigmaM ** 2)))
    D = b * A * (sigma - deltaSigma)

    return round(D)
